import argparse
import os
import logging
import sys
import csv
import platform
import shutil

import numpy as np

from time import time
from itertools import takewhile
from operator import itemgetter

from parebrick.characters.balanced import get_characters_balanced, write_characters_csv_balanced, get_characters_stats_balanced, \
    write_trees_balanced, write_stats_csv_balanced
from parebrick.characters.unbalanced import get_characters_stats_unbalanced, write_stats_csv_unbalanced, \
    write_characters_csv_unbalanced, call_unique_characters, write_trees_unbalanced
from parebrick.characters.neighbours import write_trees_neightbours
from parebrick.characters.which_chromosome import get_characters_which_chr, get_characters_stats_which_chr, \
    write_stats_csv_which_chr, write_characters_csv_which_chr, write_trees_which_chr

from parebrick.clustering.clustering import clustering, split_by_cluster

from parebrick.utils.data.converters import block_coords_to_infercars
from parebrick.utils.data.parsers import genome_lengths_from_block_coords, parse_infercars_to_df, \
    get_genomes_contain_blocks_grimm, make_labels_dict, get_block_neighbours, export_df_to_infercars, \
    genome_genome_lengths_from_chromosomes_lengths
from parebrick.utils.data.unique_gene_filters import grimm_filter_unique_gene, filter_dataframe_unique, \
    filter_dataframe_allowed
from parebrick.utils.data.stats import distance_between_blocks_dict, check_stats_stains, get_mean_coverage

from parebrick.utils.decorators import decorate

from parebrick.tree.tree_holder import TreeHolder

logger = logging.getLogger()


def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')

def restricted_float(x):
    try:
        x = float(x)
    except ValueError:
        raise argparse.ArgumentTypeError("%r not a floating-point literal" % (x,))

    if x < 0.0 or x > 1.0:
        raise argparse.ArgumentTypeError("%r not in range [0.0, 1.0]"%(x,))
    return x


# argument parsing
def initialize():
    global parser, GRIMM_FILENAME, UNIQUE_GRIMM_FILENAME, BLOCKS_COORD_FILENAME, INFERCARS_FILENAME, STATS_FILE, \
        BALANCED_FOLDER, UNBALANCED_FOLDER, CHARACTERS_FOLDER, TREES_FOLDER, BALANCED_COLORS, UNBALANCED_COLORS, \
        clustering_proximity_percentile, clustering_threshold, clustering_j, clustering_j, clustering_b, \
        CSV_BLOCK_FILENAME, CSV_BLOCK_UNIQUE_FILENAME, CSV_GENOME_LENGTH, have_unique, NEIGHBOURS_FOLDER, \
        INFERCARS_UNIQUE_FILENAME, WHICH_CHR_FOLDER

    have_unique = True

    if platform.system() == 'Linux':
        os.environ['QT_QPA_PLATFORM'] = 'offscreen'

    parser = argparse.ArgumentParser(
        description='Based on synteny blocks and phylogenetic tree this tool calls parallel rearrangements '
                    '(not consistent with phylogenetic tree).')

    required = parser.add_argument_group('Required arguments')
    optional = parser.add_argument_group('Optional arguments')

    required.add_argument('--tree', '-t', required=True, help='Tree in newick format, must be parsable by ete3 library.'
                                                              'You can read more about formats supported by ete3 at http://etetoolkit.org/docs/latest/tutorial/tutorial_trees.html#reading-and-writing-newick-trees')
    required.add_argument('--blocks_folder', '-b', required=True,
                          help='Path to folder with blocks resulted as output of original Sibelia or maf2synteny tool.')

    optional.add_argument('--output', '-o', default='parebrick_output',
                          help='Path to output folder. Default: parebrick_output.')
    optional.add_argument('--labels', '-l', default='',
                          help='Path to csv file with tree labels, must contain two columns: `strain` and `label`.')

    optional.add_argument('--show_branch_support', '-sbs', type=str2bool, default=False, const=True, nargs='?',
                          help='Show branch support while tree rendering (ete3 parameter). Default: False.')

    optional.add_argument('--keep_non_parallel', '-knp', type=str2bool, default=True,
                          help='Keep rearrangements that are not parallel in result (consistent with phylogenetic tree). Default: True.')

    optional.add_argument('--filter_for_balanced', '-fb', type=float, default=80,
                          help='Minimal percentage of block occurrences in genomes for balanced rearrangements. '
                               'All blocks with lower occurrences rate will be removed. Default: 80.')

    optional.add_argument('--visualize_neighbours', '-vn', type=str2bool, default=True,
                          help='Use module for visualizing neighbours. Default: True.')


    optional.add_argument('--which_chr', '-chr', type=str2bool, default=False,
                          help='Use information about in which chromosome block is located. Default: False.')

    optional.add_argument('--clustering_tree_patterns_coef', '-j', type=restricted_float, metavar='[0-1]', default=0.8,
                          help='Coefficient (0-1) of weight of tree patterns similarity in clustering in unbalanced (copy number variation) module.'
                               'Rest of weight will be used in distance between blocks.'
                               'E.g. if set to 0.8 (default) distance between blocks coefficient will be set to 0.2.')

    optional.add_argument('--clustering_threshold', '-c', type=restricted_float, metavar='[0-1]', default=0.025,
                          help='Threshold for algorithm of clustering, default is 0.025.'
                               'Can be increased for getting larger clusters or decreased for getting smaller and more grouped clusters.')

    clustering_proximity_percentile = 25

    GRIMM_FILENAME = 'genomes_permutations.txt'
    UNIQUE_GRIMM_FILENAME = 'genomes_permutations_unique.txt'

    BLOCKS_COORD_FILENAME = 'blocks_coords.txt'
    INFERCARS_FILENAME = 'blocks_coords.infercars'
    INFERCARS_UNIQUE_FILENAME = 'blocks_unique_coords.infercars'

    CSV_BLOCK_FILENAME = 'blocks_coords.csv'
    CSV_BLOCK_UNIQUE_FILENAME = 'blocks_unique_coords.csv'
    CSV_GENOME_LENGTH = 'genomes_lengths.csv'

    STATS_FILE = 'stats.csv'
    BALANCED_FOLDER = 'balanced_rearrangements_output/'
    UNBALANCED_FOLDER = 'un' + BALANCED_FOLDER
    NEIGHBOURS_FOLDER = 'neighbours_output/'
    WHICH_CHR_FOLDER = 'which_chromosome_output/'

    CHARACTERS_FOLDER = 'characters/'
    TREES_FOLDER = 'tree_colorings/'

    BALANCED_COLORS = ['White', 'Gainsboro', 'DimGray', 'LightGreen', 'LightBlue', 'NavajoWhite', 'LightPink',
                       'DarkSeaGreen', 'Orchid', 'Navy', 'Olive', 'Teal', 'SaddleBrown', 'SeaGreen', 'DarkCyan',
                       'DarkOliveGreen', 'LightCoral']

    UNBALANCED_COLORS = ['Gainsboro', 'White'] + BALANCED_COLORS[3:]


# This module converts data from the output data format
# of Sibelia or Ragout scripts into the infercars format to simplify the subsequent annotation.
# Also filtering blocks in the grimm format for unique single-copy blocks for the breakpoint graph construction.
@decorate("Preprocess Data", logger)
def preprocess_data():
    global unique_blocks, balanced_block_rate
    unique_blocks = grimm_filter_unique_gene(blocks_folder + GRIMM_FILENAME,
                                             preprocessed_data_folder + UNIQUE_GRIMM_FILENAME, balanced_block_rate)
    logger.info('Converting block coords to infercars format')
    block_coords_to_infercars(blocks_folder + BLOCKS_COORD_FILENAME, preprocessed_data_folder + INFERCARS_FILENAME)


# In this module, parsing of input blocks and their coordinates, as well as a phylogenetic, takes place.
# Next, some statistics are calculated: the total number of blocks,
# the number of unique single-copy blocks, as well as the percentage of genome coverage with these two types of blocks.
# After that, the correspondence of strains set on the phylogenetic tree and strains set in the synthenic blocks
# is checked, if necessary, the missing strains are discarded.
@decorate("Parsers and check strains", logger)
def parsers_and_stats():
    global chr_lengths, blocks_df, tree_holder, genomes, blocks, block_genome_count, have_unique, unique_blocks, \
        neighbours

    chr_lengths = genome_lengths_from_block_coords(blocks_folder + BLOCKS_COORD_FILENAME)
    genome_lengths = genome_genome_lengths_from_chromosomes_lengths(chr_lengths)

    blocks_df = parse_infercars_to_df(preprocessed_data_folder + INFERCARS_FILENAME)
    unique_blocks_df = filter_dataframe_unique(blocks_df)
    filted_blocks_df = filter_dataframe_allowed(blocks_df, unique_blocks)
    neighbours = get_block_neighbours(blocks_folder + GRIMM_FILENAME)

    if len(filted_blocks_df) == 0:
        have_unique = False
        logger.warning('No unique one-copy blocks found. Balanced rearrangements will not be called')

    export_df_to_infercars(unique_blocks_df, preprocessed_data_folder + INFERCARS_UNIQUE_FILENAME)
    blocks_df.to_csv(preprocessed_data_folder + CSV_BLOCK_FILENAME, index=False)
    unique_blocks_df.to_csv(preprocessed_data_folder + CSV_BLOCK_UNIQUE_FILENAME, index=False)
    with open(preprocessed_data_folder + CSV_GENOME_LENGTH, 'w') as f:
        w = csv.writer(f)
        w.writerow(['Genome', 'Length'])
        w.writerows(genome_lengths.items())

    logger.info(f'Blocks count: {len(blocks_df["block"].unique())}')
    logger.info(f'Unique one-copy blocks count: {len(unique_blocks_df["block"].unique())}')

    logger.info(f'Mean block coverage: {get_mean_coverage(blocks_df, genome_lengths) * 100} %')
    logger.info(f'Mean common one-copy blocks coverage: '
                f'{get_mean_coverage(unique_blocks_df, genome_lengths) * 100 if have_unique else 0} %')
    logger.info(f'Mean {balanced_block_rate} % rate one-copy blocks coverage: '
                f'{get_mean_coverage(filted_blocks_df, genome_lengths) * 100 if have_unique else 0} %')

    tree_holder = TreeHolder(tree_file, logger, labels_dict=make_labels_dict(labels_file))

    genomes, blocks, block_genome_count = get_genomes_contain_blocks_grimm(blocks_folder + GRIMM_FILENAME)

    genomes = check_stats_stains(tree_holder, set(genomes), logger)


# In this module, the breakpoint of the graph is built using the bg library,
# then an algorithm for constructing characters and their states for each edge of the breakpoint graph is implemented.
@decorate("Balanced rearrangements characters", logger)
def balanced_rearrangements_characters():
    global b_characters
    b_characters = get_characters_balanced(preprocessed_data_folder + UNIQUE_GRIMM_FILENAME, genomes, logger)


# This module implements balanced rearrangements characters statistics calculation
# (measure for non-convexity proposed in the paper),
# as well as an estimate of the average distance of the length of the breakdown in the genome in nucleotides.
@decorate("Balanced rearrangements stats", logger)
def balanced_rearrangements_stats():
    global b_characters, b_stats, keep_consistent

    logger.info('Counting distances between unique one-copy blocks, may take a while')
    distance_between_uniq_blocks = distance_between_blocks_dict(blocks_df, chr_lengths, unique_blocks)
    b_stats = get_characters_stats_balanced(b_characters, tree_holder, distance_between_uniq_blocks)
    char_stats = zip(b_characters, b_stats)

    logger.info(f'Got characters after breakpoint graph consideration: {len(b_characters)}')
    # sorting for get most interesting characters in the beginning
    # [1][3] for vertex1,vertex2,parallel_rear_score, [1][6] for parallel_breakpoint_score, r[0][0] for vertex1
    char_stats = sorted(char_stats, key=lambda r: (r[1][2], r[1][5], r[0][0]), reverse=True)
    # remove convex characters
    if not keep_consistent:
        char_stats = list(takewhile(lambda r: not r[1][7], char_stats))
        logger.info(f'Left non-convex characters after filtering: {len(char_stats)}')
    else:
        logger.info('Percentage of consistent rearrangements (balanced): '
                    f'{np.mean([r[1][7] for r in char_stats]) * 100} %')

    # unzip
    b_characters = [char for char, stat in char_stats]
    b_stats = [stat for char, stat in char_stats]


# This module generates the output method files: the file stats.csv with the statistics
# of balanced rearrangements of all non-convex features.
# The tree_colorings folder contains characters visualized on the phylogenetic tree in .pdf format,
# different colors correspond to different states of the character, and the characters folder contains the .csv format
# files for the character states in different strains for each character.
@decorate("Balanced rearrangements output", logger)
def balanced_rearrangements_output():
    balanced_folder = output_folder + BALANCED_FOLDER
    os.makedirs(balanced_folder, exist_ok=True)

    stats_file = balanced_folder + STATS_FILE
    write_stats_csv_balanced(b_stats, stats_file)

    characters_folder = balanced_folder + CHARACTERS_FOLDER
    write_characters_csv_balanced(b_characters, characters_folder)

    trees_folder = balanced_folder + TREES_FOLDER
    write_trees_balanced(b_characters, trees_folder, show_branch_support, tree_holder, BALANCED_COLORS)


# In this module, the mapping of blocks by its number for each strain is performed.
@decorate("Unbalanced rearrangements characters", logger)
def unbalanced_rearrangements_characters():
    global ub_characters
    ub_characters = [{genome: block_genome_count[block][genome] for genome in genomes}
                     for block in blocks]


# This module implements balanced rearrangements characters statistics calculation
# (measure for non-convexity proposed in the paper),
# as well as the construction of distance matrices based on a measure of Jacquard similarity and distance in nucleotides
# and the subsequent clustering for unbalanced characters.
@decorate("Unbalanced rearrangements stats and clustering", logger)
def unbalanced_rearrangements_stats_and_clustering():
    global ub_cls, ub_characters, ub_stats
    ub_stats = get_characters_stats_unbalanced(blocks, ub_characters, tree_holder)

    char_stats = zip(ub_characters, ub_stats)
    logger.info(f'Got characters after copy number variation consideration: {len(blocks)}')
    # sorting for get most interesting characters in the beginning
    # [1][3] for vertex1,vertex2,parallel_rear_score, [1][6] for parallel_breakpoint_score, r[0][0] for vertex1
    char_stats = sorted(char_stats, key=lambda r: (r[1][1], r[1][0]), reverse=True)
    # remove convex characters
    if not keep_consistent:
        char_stats = list(takewhile(lambda r: not r[1][4], char_stats))
        logger.info(f'Left non-convex characters after filtering: {len(char_stats)}')
    else:
        logger.info('Percentage of consistent rearrangements (unbalanced): '
                    f'{np.mean([r[1][4] for r in char_stats]) * 100} %')

    # unzip
    ub_characters = [char for char, stat in char_stats]
    ub_stats = [stat for char, stat in char_stats]

    if len(ub_stats) == 0:
        logger.info('There is no non-convex characters, skipping clustering part')
        return
    logger.info('Counting distances between non-convex character blocks, may take a while')
    if len(ub_stats) > 1:
        distance_between_blocks = distance_between_blocks_dict(blocks_df, chr_lengths,
                                                               set(map(itemgetter(0), ub_stats)))
        ub_cls = clustering(ub_characters, ub_stats, distance_between_blocks, max(chr_lengths.values()),
                            clustering_threshold, clustering_j, clustering_b, clustering_proximity_percentile, logger)
    else:
        ub_cls = np.array([0])

    logger.info(f'Clusters: {np.unique(ub_cls).shape[0]}')


# In this module, the output of the method files is generated: the file stats.csv with the statistics of unbalanced
# genomic rearrangements of all non-convex characters.
# The tree_colorings folder contains characters visualized on the phylogenetic tree in .pdf format,
# and in the characters folder contain the .csv files for the character states in different strains
# for all the characters.
# These .pdf and .csv files are located in subfolders according to their presentation in clustering.
@decorate('Unbalanced rearrangements output', logger)
def unbalanced_rearrangements_output():
    unbalanced_folder = output_folder + UNBALANCED_FOLDER
    os.makedirs(unbalanced_folder, exist_ok=True)

    stats_file = unbalanced_folder + STATS_FILE
    write_stats_csv_unbalanced(ub_stats, ub_cls, stats_file)

    characters_folder = unbalanced_folder + CHARACTERS_FOLDER

    cls_chars, cls_stats = split_by_cluster(ub_characters, ub_stats, ub_cls)
    unique_chars_list = call_unique_characters(cls_chars, cls_stats)

    write_characters_csv_unbalanced(unique_chars_list, characters_folder)

    trees_folder = unbalanced_folder + TREES_FOLDER
    write_trees_unbalanced(unique_chars_list, trees_folder, show_branch_support, tree_holder, UNBALANCED_COLORS)


@decorate('Visualize neighbours output', logger)
def neighbours_output():
    neighbours_folder = output_folder + NEIGHBOURS_FOLDER
    os.makedirs(neighbours_folder, exist_ok=True)

    write_trees_neightbours([s[0] for s in ub_stats], ub_characters, neighbours, neighbours_folder, show_branch_support,
                            tree_holder, UNBALANCED_COLORS)


@decorate("Which chromosome characters", logger)
def which_chromosome_characters():
    global chr_characters
    chr_characters = get_characters_which_chr(blocks_df)


@decorate("Which chromosome stats and clustering", logger)
def which_chromosome_stats_and_clustering():
    global chr_cls, chr_characters, chr_stats
    chr_stats = get_characters_stats_which_chr(chr_characters, tree_holder)

    char_stats = zip(chr_characters, chr_stats)
    logger.info(f'Got characters after which chromosome consideration: {len(blocks)}')
    # sorting for get most interesting characters in the beginning
    char_stats = sorted(char_stats, key=lambda r: (r[1][4], r[1][3]), reverse=True)
    # remove convex characters
    if not keep_consistent:
        char_stats = list(takewhile(lambda r: not r[1][7], char_stats))
        logger.info(f'Left non-convex characters after filtering: {len(char_stats)}')
    else:
        logger.info('Percentage of consistent rearrangements (unbalanced): '
                    f'{np.mean([r[1][7] for r in char_stats]) * 100} %')

    # unzip
    chr_characters = [char for char, stat in char_stats]
    chr_stats = [stat for char, stat in char_stats]


@decorate('Which chromosome output', logger)
def which_chromosome_output():
    chr_folder = output_folder + WHICH_CHR_FOLDER
    os.makedirs(chr_folder, exist_ok=True)

    stats_file = chr_folder + STATS_FILE
    write_stats_csv_which_chr(chr_stats, stats_file)

    characters_folder = chr_folder + CHARACTERS_FOLDER

    write_characters_csv_which_chr(chr_characters, characters_folder)

    trees_folder = chr_folder + TREES_FOLDER
    write_trees_which_chr(chr_characters, trees_folder, show_branch_support, tree_holder, UNBALANCED_COLORS)


def main():
    def logger_initialize():
        file_handler = logging.FileHandler(filename=output_folder + 'run.log')
        stdout_handler = logging.StreamHandler(sys.stdout)
        handlers = [file_handler, stdout_handler]

        logging.basicConfig(
            level=logging.DEBUG,
            format='[%(asctime)s] %(levelname)s - %(message)s',
            handlers=handlers
        )

    global blocks_folder, output_folder, tree_file, labels_file, preprocessed_data_folder, show_branch_support, \
        have_unique, keep_consistent, balanced_block_rate, clustering_threshold, clustering_j, clustering_b
    initialize()

    start_time = time()
    d = vars(parser.parse_args())
    blocks_folder, output_folder, tree_file, labels_file, show_branch_support, keep_consistent, balanced_block_rate, \
    visualize_neighbours, clustering_j, clustering_threshold, which_chr_flag = d['blocks_folder'], d['output'], \
           d['tree'], d['labels'], d['show_branch_support'], d['keep_non_parallel'], d['filter_for_balanced'], \
           d['visualize_neighbours'], d['clustering_tree_patterns_coef'], d['clustering_threshold'], d['which_chr']

    clustering_b = 1 - clustering_j

    # folders
    if blocks_folder[-1] != '/': blocks_folder += '/'
    if output_folder[-1] != '/': output_folder += '/'
    shutil.rmtree(output_folder, ignore_errors=True)
    print('Cleaning output folder')

    preprocessed_data_folder = output_folder + 'preprocessed_data/'
    os.makedirs(preprocessed_data_folder, exist_ok=True)

    logger_initialize()

    preprocess_data()
    parsers_and_stats()

    if have_unique:
        balanced_rearrangements_characters()
        balanced_rearrangements_stats()
        balanced_rearrangements_output()

    unbalanced_rearrangements_characters()
    unbalanced_rearrangements_stats_and_clustering()
    if len(ub_stats) != 0: unbalanced_rearrangements_output()

    if which_chr_flag:
        which_chromosome_characters()
        which_chromosome_stats_and_clustering()
        which_chromosome_output()

    if visualize_neighbours:
        neighbours_output()

    logger.info(f'Total elapsed time: {time() - start_time} seconds')


if __name__ == "__main__":
    main()



from ete3 import SeqMotifFace
from collections import defaultdict, Counter
from itertools import chain

import numpy as np


def get_neighbour_motifs(n, ns_colors, offset, inverse=False):
    block, orient = n[:-1], n[-1]
    clr = ns_colors[block]

    if (orient == 'h' and not inverse) or (inverse and orient == 't'):
        return [[offset, offset + 30, "[]", None, 10, clr, clr, f"arial|2|white|+{block}"],
                [offset + 30, offset + 40, ">", None, 10, clr, clr, None]]
    else:
        return [[offset, offset + 10, "<", None, 10, clr, clr, None],
                [offset + 10, offset + 40, "[]", None, 10, clr, clr, f"arial|2|white|-{block}"]]


def generate_neighbour_face(node_ns, ns_colors, block, offsets):
    motifs = [[0, 0, "blank", None, 10, None, None, None]]

    for cur_offset, next_offset, node_n in zip(offsets, offsets.tolist()[1:], node_ns):
        if len(node_n) > 0:
            (nbr1, nbr2, copies, orientations) = node_n

            motifs.extend(get_neighbour_motifs(nbr1, ns_colors, cur_offset))

            for i, or_ in enumerate(orientations):
                motifs.extend(
                    get_neighbour_motifs(f'{block}' + ('h' if or_ == '+' else 't'), ns_colors, cur_offset + (i + 1) * 50))

            motifs.extend(get_neighbour_motifs(nbr2, ns_colors, cur_offset + (copies + 1) * 50, True))

            motifs.append([cur_offset + (copies + 1) * 50 + 40, next_offset, "blank", None, 10, None, None, None])
        else:
            motifs.append([cur_offset, next_offset, "blank", None, 10, None, None, None])

    for offset in offsets[1:-1]:
        motifs.append([offset - 10, offset - 10, "[]", None, 10, 'grey', 'grey', None])
        motifs.append([offset - 10, offset, "blank", None, 10, None, None, None])

    return SeqMotifFace('', motifs=motifs)


def align_neighbours(neighbours, all_genomes):
    aligned_neighbours = {}

    sorted_neighbours_keys = sorted(all_genomes, key=lambda k: len(neighbours[k]), reverse=True)
    max_blocks = len(neighbours[sorted_neighbours_keys[0]])

    l_neighbour_to_position = defaultdict(Counter)
    r_neighbour_to_position = defaultdict(Counter)

    align_smart = False
    for genome in sorted_neighbours_keys:
        node_ns = neighbours[genome]

        node_ns = sorted(node_ns, key=lambda k: (int(k[0][:-1]), int(k[1][:-1])))

        n = len(node_ns)
        # setting new indexes to better alignment
        if align_smart:
            new_node_ns = [()] * max_blocks
            used_nodes = [False for _ in range(n)]
            used_indexes = [False for _ in range(max_blocks)]

            for i, node_n in enumerate(node_ns):
                n1, n2, _1, _2 = node_n

                for new_ind, _freq in sorted(chain(l_neighbour_to_position[n1[:-1]].most_common(),
                                                   map(lambda v: (v[0], v[1] * 0.9),
                                                       r_neighbour_to_position[n2[:-1]].most_common())),
                                             key=lambda v: v[1], reverse=True):
                    if not used_indexes[new_ind]:
                        new_node_ns[new_ind] = node_n
                        used_nodes[i] = True
                        used_indexes[new_ind] = True
                        break

            free_old_index = [i for i, used in enumerate(used_nodes) if not used]
            free_new_index = [i for i, used in enumerate(used_indexes) if not used]

            assert len(free_old_index) <= len(free_new_index)

            for oi, ni in zip(free_old_index, free_new_index):
                new_node_ns[ni] = node_ns[oi]

            node_ns = new_node_ns

        # remember positions
        for i, node_n in enumerate(node_ns):
            if len(node_n) > 0:
                (n1, n2, _1, _2) = node_n
                l_neighbour_to_position[n1[:-1]][i] += 1
                r_neighbour_to_position[n2[:-1]][i] += 1

        align_smart = True
        aligned_neighbours[genome] = node_ns

    return aligned_neighbours


def get_offsets(neighbours):
    copies_2d = [[node_n[2] if len(node_n) > 0 else 0 for node_n in node_ns]
                 for node_ns in neighbours.values()]

    copies = np.max(copies_2d, axis=0)
    basic_offset = np.arange(0, 160 * len(copies) + 1, 160)

    shifted_copies = np.ones(len(copies) + 1, dtype='i4')
    shifted_copies[1:] = copies

    return basic_offset + np.cumsum((shifted_copies - 1) * 50)



import time

sep = '-' * 80

def decorate(module_name, logger):
    def actual_decorator(func):

        def wrapper(*args, **kwargs):
            start = time.time()

            logger.info(f'<Running module: {module_name}>')
            print(sep)
            return_value = func(*args, **kwargs)

            end = time.time()

            print(sep)
            logger.info(f'<Module finished: {module_name}>, elapsed {end - start} seconds')
            print()
            print()
            return return_value

        return wrapper

    return actual_decorator

# def print_module_name():



from parebrick.utils.data.parsers import parse_infercars_to_df
from parebrick.utils.data.stats import distance_between_blocks_distribution
from parebrick.utils.data.unique_gene_filters import filter_dataframe_unique

import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np

import argparse
import os


def blocks_length_dist(df):
    return [end_ - start_ for start_, end_ in zip(df['chr_beg'], df['chr_end'])]


def lengths_between(state, log):
    plt.figure()
    if state == 'after':
        df_filtered = filter_dataframe_unique(df)
        ds = distance_between_blocks_distribution(df_filtered)
    else:
        ds = distance_between_blocks_distribution(df)

    sns.histplot(ds, bins=100, log_scale=(False, log), kde_kws={'alpha': 0.7})

    plt.ylabel('Number of blocks')
    plt.xlabel('Length in nucleotides')
    plt.title(f'Length of fragments not covered by {"common" if state == "after" else "any"} blocks')

    plt.tight_layout()
    plt.xlim(xmin=0)

    plt.savefig(out_folder + f'lengths_between_{state}_filtering{"_log" if log else ""}.pdf')
    # plt.show()


def number_of_genomes_weighted(weighted, log):
    plt.figure()
    vs, ws = [], []

    for _, df_block in df.groupby('block'):
        vs.append(len(df_block.species.unique()))
        lens = [row['chr_end'] - row['chr_beg'] for _, row in df_block.iterrows()]
        ws.append(np.mean(lens))

    bins = 100 if max(vs) > 100 else max(vs)
    if weighted:
        plt.hist(vs, bins=bins, weights=ws, log=log, alpha=0.7)
    else:
        sns.histplot(vs, bins=bins, log_scale=(False, log), kde_kws={'alpha': 0.7})

    plt.ylabel('Length of fragments that are present\n in n genomes, nucleotides'
               if weighted else 'Number of blocks')
    plt.xlabel('Number of genomes')
    plt.title(f'{"Weighted f" if weighted else "F"}requency of synteny blocks')

    plt.xlim(xmin=0, xmax=max(vs))
    plt.tight_layout()
    plt.savefig(out_folder + f'blocks_frequency{"_weighted" if weighted else ""}{"_log" if log else ""}.pdf')
    # plt.show()


def block_length(log):
    plt.figure()
    df_filtered = filter_dataframe_unique(df)
    ds_before = blocks_length_dist(df)
    ds_after = blocks_length_dist(df_filtered)

    bins = np.linspace(min(ds_before), max(ds_before), 100)
    sns.histplot(ds_before, bins=bins, log_scale=(False, log), kde_kws={'alpha': 0.7}, label='not-common')
    sns.histplot(ds_after, bins=bins, log_scale=(False, log), kde_kws={'alpha': 0.7}, label='common', color='red')

    plt.ylabel('Number of blocks')
    plt.xlabel('Length in nucleotides')
    plt.xlim(xmin=0)
    plt.title(f'Distribution of synteny blocks length')

    plt.legend(loc=1)

    plt.tight_layout()
    plt.savefig(out_folder + f'block_lengths_distribution{"_log" if log else ""}.pdf')
    # plt.show()


def scatter_len_genomes_count(log):
    plt.figure()
    xs = []
    ys = []
    for block, df_block in df.groupby('block'):
        xs.append(len(df_block.species.unique()))
        lens = [row['chr_end'] - row['chr_beg'] for _, row in df_block.iterrows()]
        ys.append(np.mean(lens))

    if log: plt.yscale('log')
    plt.scatter(xs, ys, s=1)

    plt.xlabel('Number of genomes')
    plt.ylabel('Length of blocks')
    plt.title(f'Occurrence of synteny blocks vs its length')

    plt.xlim(xmin=0 - 0.02 * (max(xs)), xmax=max(xs) + 0.02 * (max(xs)))
    if not log: plt.ylim(ymin=0)

    plt.tight_layout()
    plt.savefig(out_folder + f'scatter_number_length{"_log" if log else ""}.pdf')

    plt.subplots_adjust(left=0.14, right=0.99)

    # plt.show()


def pan_blocks(permutations=10000):
    block_sets = [set(df_sp.block.unique()) for _, df_sp in df.groupby('species')]
    nbss, pnss = [], []
    for _ in range(permutations):
        block_sets = np.random.permutation(block_sets)
        nbs, pns = [], []
        accumulate_set = set()
        for bs in block_sets:
            left = bs - accumulate_set
            nbs.append(len(left))
            accumulate_set |= bs
            pns.append(len(accumulate_set))

        nbss.append(nbs)
        pnss.append(pns)

    nbss = np.array(nbss)
    xs = list(range(1, len(block_sets) + 1))

    def new_blocks():
        plt.figure()
        plt.plot(xs, np.median(nbss, axis=0), label='median (different permutations)')
        plt.fill_between(xs, np.percentile(nbss, 5, axis=0), np.percentile(nbss, 95, axis=0), alpha=0.4,
                         label='90% confidence interval')

        plt.xlabel('number of genomes')
        plt.ylabel('new blocks')
        plt.title('New blocks as a function of number of genomes')
        plt.legend(loc='upper right')

        plt.ylim(ymax=np.percentile(nbss, 95, axis=0)[1], ymin=0)
        plt.xlim(xmin=0, xmax=max(xs))

        # plt.subplots_adjust(top=0.99, right=0.99)
        plt.tight_layout()
        plt.savefig(out_folder + 'new_blocks.pdf')
        # plt.show()

    def pan():
        plt.figure()
        plt.plot(xs, np.median(pnss, axis=0), label='median (different permutations)', color='indianred')
        plt.fill_between(xs, np.percentile(pnss, 5, axis=0), np.percentile(pnss, 95, axis=0), alpha=0.4,
                         label='90% confidence interval', color='indianred')

        plt.xlabel('number of genomes')
        plt.ylabel('Pan-blocks count')
        plt.title('Pangenome')
        plt.legend(loc='lower right')

        plt.ylim(ymin=np.percentile(pnss, 5, axis=0)[0])
        plt.xlim(xmin=0, xmax=max(xs))
        # plt.subplots_adjust(top=0.99, right=0.99)
        plt.tight_layout()
        plt.savefig(out_folder + 'pan_blocks.pdf')
        # plt.show()

    new_blocks()
    pan()

def main():
    global out_folder, df

    parser = argparse.ArgumentParser(
        description='Building charts for pan-genome analysis based on synteny blocks.')

    parser.add_argument('--infercars_file', '-f', required=True,
                      help='Path to file in infercars format, can be found in main script output')

    parser.add_argument('--output', '-o', default='parebrick_charts', help='Path to output folder.')

    args = parser.parse_args()
    d = vars(args)

    file, out_folder = d['infercars_file'], d['output']

    if out_folder[-1] != '/': out_folder += '/'
    os.makedirs(out_folder, exist_ok=True)

    df = parse_infercars_to_df(file)
    sns.set(style="whitegrid", font="serif")

    print('Plotting lengths between blocks')
    lengths_between('before', log=False)
    lengths_between('after', log=False)
    lengths_between('before', log=True)
    lengths_between('after', log=True)

    print('Plotting number of genomes in blocks')
    number_of_genomes_weighted(weighted=False, log=False)
    number_of_genomes_weighted(weighted=True, log=False)
    number_of_genomes_weighted(weighted=False, log=True)
    number_of_genomes_weighted(weighted=True, log=True)

    print('Plotting blocks length distribution')
    block_length(log=True)
    block_length(log=False)

    print('Plotting scatter for occurrence of synteny blocks vs its length')
    scatter_len_genomes_count(False)
    scatter_len_genomes_count(True)

    print('Plotting pan-genome plots')
    pan_blocks()

if __name__ == "__main__":
    main()


from bg.grimm import GRIMMReader
from collections import Counter, defaultdict


class Unique_Filter:
    def __init__(self):
        self.blocks_copies = defaultdict(lambda: defaultdict(int))
        self.first_call = True

    def update_allowed_blocks(self, ps, strain):
        vs = [p[1] for p in ps]
        for v in vs:
            self.blocks_copies[v][strain] += 1

    def count_allowed(self, min_genomes):
        self.allowed_blocks = [b for b, cs in self.blocks_copies.items()
                               if all(map(lambda c: c == 1, cs.values())) and len(cs) >= min_genomes]

    def filter_unique(self, ps):
        return [p for p in ps if p[1] in self.allowed_blocks]

def grimm_filter_unique_gene(in_file, out_file, block_rate):
    lines = open(in_file).read().split('\n')
    strains = set()

    # make unique blocks list
    i = 0
    flt = Unique_Filter()
    while i < len(lines):
        line = lines[i]
        if GRIMMReader.is_genome_declaration_string(line):
            strain = GRIMMReader.parse_genome_declaration_string(line).name
            strains.add(strain)

            data_line = lines[i + 1]
            parsed = GRIMMReader.parse_data_string(data_line)[1]
            flt.update_allowed_blocks(parsed, strain)
            i += 2
        else:
            i += 1

    flt.count_allowed(block_rate * len(strains) / 100)
    # write allowed blocks
    i = 0
    with open(out_file, 'w') as f:
        while i < len(lines):
            line = lines[i]
            if GRIMMReader.is_genome_declaration_string(line):
                data_line = lines[i + 1]

                parsed = GRIMMReader.parse_data_string(data_line)[1]
                parsed = flt.filter_unique(parsed)

                print(line, file=f)
                print(' '.join(p[0] + p[1] for p in parsed), '@', file=f)
                i += 2
            else:
                i += 1

    return list(map(int, flt.allowed_blocks))

def filter_dataframe_unique(df):
    allowed_blocks = set()
    all_sp = len(df['species'].unique())
    for block, df_block in df.groupby('block'):
        if len(df_block) == len(df_block['species'].unique()) == all_sp:
            allowed_blocks.add(block)

    return df.loc[df['block'].isin(allowed_blocks)].copy()

def filter_dataframe_allowed(df, allowed_blocks):
    return df.loc[df['block'].isin(allowed_blocks)].copy()


import numpy as np

from parebrick.clustering.distance_matrices import dist_matrix_similarity, dist_matrix_proximity

from sklearn.cluster import AgglomerativeClustering


def renew_index(cls):
    used, new_cls_indexes = set(), []
    for cl in cls:
        if not cl in used:
            used.add(cl)
            new_cls_indexes.append(cl)

    new_index_dict = {cl: i for i, cl in enumerate(new_cls_indexes)}
    return np.array([new_index_dict[cl] for cl in cls])

def clustering(characters, stats, distance_between_blocks, max_length, threshold, j, b, proximity_percentile, logger):
    J = dist_matrix_similarity(characters)
    if J.max() != 0: J /= J.max()
    logger.info('Jaccard index matrix constructed')

    B = dist_matrix_proximity(stats, distance_between_blocks, max_length, proximity_percentile)
    B /= B.max()
    logger.info('Proximity matrix constructed')

    D = J * j + B * b
    cls = AgglomerativeClustering(n_clusters=None, metric='precomputed', linkage='average',
                                  distance_threshold=threshold).fit_predict(D)

    logger.info('Clustring is done')

    return renew_index(cls)


def split_by_cluster(characters, stats, cls):
    cls_chars, cls_stats = [], []
    for cl in np.unique(cls):
        inds = np.where(cls == cl)[0]
        cur_characters = [characters[i] for i in inds]
        cur_stats = [stats[i] for i in inds]

        cls_chars.append(cur_characters)
        cls_stats.append(cur_stats)
    return cls_chars, cls_stats


import os
import re

import numpy as np
import pandas as pd

from itertools import takewhile
from io import StringIO
from bg.grimm import GRIMMReader
from collections import defaultdict, Counter

PATTERN = re.compile("([A-Za-z0-9_\(\)\/\s\.-]+)\.([A-Za-z0-9_]+):(\d+)-(\d+) ([+|-]).*")
COLUMNS = ["block", "species", "chr", "chr_beg", "chr_end", "orientation"]
BLOCKS_SEPARATOR = '-' * 80


def parse_infercars_to_df(file_name):
    def find_indices(lst, condition):
        return [i for i, elem in enumerate(lst) if condition(elem)]

    with open(file_name) as f:
        lines = f.readlines()

    last_line = len(lines) - 1
    while lines[last_line] == '\n': last_line -= 1

    n_at_end = len(lines) - 1 - last_line
    for _ in range(1 - n_at_end): lines.append('\n')

    bs = np.split(lines, find_indices(lines, lambda x: x[0] == ">"))
    temp = []

    for i, b in enumerate(bs):
        if len(b) == 0: continue
        b_i = int(b[0][1:])

        for oc in b[1:-1]:
            m = PATTERN.match(oc)
            temp.append([b_i, m.group(1), m.group(2), int(m.group(3)), int(m.group(4)), m.group(5)])

    return pd.DataFrame(temp, columns=COLUMNS)


def export_df_to_infercars(df, file_name):
    with open(file_name, 'w') as f:
        for block, block_df in df.groupby('block'):
            print(f'>{block}', file=f)
            for i, row in block_df.iterrows():
                print(f'{row["species"]}.{row["chr"]}:{row["chr_beg"]}-{row["chr_end"]} {row["orientation"]}', file=f)
            print(file=f)


def genome_lengths_from_block_coords(in_file):
    with open(in_file) as f:
        head_lines = list(takewhile(lambda line: (line != BLOCKS_SEPARATOR + os.linesep) and
                                                 (line != BLOCKS_SEPARATOR + '\n'), f))

    # names of chromosomes
    df_head = pd.read_csv(StringIO(''.join(head_lines)), sep='\t')
    return {row['Description'] + '.1': row['Size'] for index, row in df_head.iterrows()}

def genome_genome_lengths_from_chromosomes_lengths(chr_lengths):
    lengths = defaultdict(int)

    for chr, len in chr_lengths.items():
        strain, _ = chr.rsplit('.', 1)
        lengths[strain] += len

    return lengths


def get_genomes_contain_blocks_grimm(grimm_file):
    genomes, blocks = set(), set()

    with open(grimm_file) as f:
        ls = f.readlines()
    block_genome_count = defaultdict(Counter)

    for i in range(0, len(ls), 2):
        name = GRIMMReader.parse_genome_declaration_string(ls[i]).name
        data = GRIMMReader.parse_data_string(ls[i + 1])[1]
        genomes.add(name)
        for _, block in data:
            blocks.add(int(block))
            block_genome_count[int(block)][name] += 1

    return list(sorted(genomes)), list(sorted(blocks)), block_genome_count


def get_block_neighbours(grimm_file):
    block_neighbours = defaultdict(lambda: defaultdict(list))
    with open(grimm_file) as f: ls = f.readlines()

    i = 0
    while i < len(ls):
        l = ls[i]
        if GRIMMReader.is_genome_declaration_string(l):
            strain = GRIMMReader.parse_genome_declaration_string(l).name
            data_line = ls[i + 1]
            bs = GRIMMReader.parse_data_string(data_line)[1]

            n = len(bs)
            j = 0

            while j < n:
                tandem_copies = 1
                prev_or, prev_block = bs[j % n]
                _, curr_block = bs[(j + 1) % n]
                next_or, next_block = bs[(j + 2) % n]

                if curr_block == prev_block:
                    j += 1
                    continue

                while curr_block == next_block:
                    j += 1
                    tandem_copies += 1
                    next_or, next_block = bs[(j + 2) % n]

                neighbours = (prev_block + ('h' if prev_or == '+' else 't'),
                              next_block + ('t' if next_or == '+' else 'h'))

                orientations = tuple(bs[(k + 1) % n][0] for k in range(j - tandem_copies + 1, j + 1))

                if orientations[0] == '-':
                    neighbours = (neighbours[1], neighbours[0])
                    orientations = tuple('+' if or_ == '-' else '+' for or_ in orientations[::-1])

                block_neighbours[int(curr_block)][strain].append((*neighbours, tandem_copies, orientations))

                j += 1

            i += 2
        else:
            i += 1

    return block_neighbours


def make_labels_dict(file, row_from='strain', row_to='label'):
    try:
        df = pd.read_csv(file)
        return {row[row_from]: row[row_to] for i, row in df.iterrows()}
    except FileNotFoundError:
        return None


from ete3 import Tree, TreeStyle, TextFace, RectFace

from collections import defaultdict, Counter
from itertools import combinations

from parebrick.tree.neighbours_utils import generate_neighbour_face, align_neighbours, get_offsets


class TreeHolder:
    def __init__(self, tree, logger, scale=None, labels_dict=None, node_colors=defaultdict(lambda: 'black'),
                 reroot=True, hz_line_width=1, vt_line_width=1):
        self.tree = Tree(tree)
        self.scale = scale

        if reroot:
            r = self.tree.get_midpoint_outgroup()
            self.tree.set_outgroup(r)

        for node in self.tree.traverse():
            if len(node.children) == 3:
                logger.info("Trying to root tree by first child of root")
                logger.info(f'Children of root: {node.children}')
                self.tree.set_outgroup(node.children[0])
            break

        for node in self.tree.traverse():
            # Hide node circles
            node.img_style['size'] = 0
            node.img_style['hz_line_width'] = hz_line_width
            node.img_style['vt_line_width'] = vt_line_width

            if node.is_leaf():
                try:
                    name_face = TextFace(labels_dict[node.name] if labels_dict else node.name,
                                         fgcolor=node_colors[node.name])
                except KeyError:
                    msg = f'There is not label for leaf {node.name} in labels file'
                    logger.error(msg)
                    raise KeyError(msg)
                node.add_face(name_face, column=0)

    def draw_neighbours(self, neighbours, block, colors=('Crimson', 'Teal', 'DarkGreen', 'Purple', 'DarkKhaki',
                                                         'MediumVioletRed', 'DarkOrange', 'Navy', 'RosyBrown',
                                                         'DarkGoldenrod', 'Sienna', 'Indigo', 'DarkRed', 'Olive',
                                                         'SlateGray', 'SeaGreen', 'IndianRed', 'BurlyWood')):
        posible_ns = sorted(list(set(n[:-1] for nss in neighbours.values() for ns in nss for n in ns[:2])))

        ns_colors = {posible_ns[i]: colors[i % len(colors)] for i in range(len(posible_ns))}
        ns_colors[str(block)] = 'grey'

        # if block != 2: return
        all_genomes = [node.name for node in self.tree.traverse()]
        aligned_neighbours = align_neighbours(neighbours, all_genomes)
        offsets = get_offsets(aligned_neighbours)

        for node in self.tree.traverse():
            if not node.is_leaf(): continue
            face = generate_neighbour_face(aligned_neighbours[node.name], ns_colors, block, offsets)
            node.add_face(face, 1, "aligned")

    def draw(self, file, colors, color_internal_nodes=True, legend_labels=(), show_branch_support=True,
             show_scale=True, legend_scale=1, mode="c", neighbours=None, neighbours_block=None):
        max_color = len(colors)

        used_colors = set()
        for node in self.tree.traverse():
            if not (color_internal_nodes or node.is_leaf()): continue
            color = colors[min(node.color, max_color - 1)]
            node.img_style['bgcolor'] = color
            used_colors.add(color)

        ts = TreeStyle()
        ts.mode = mode
        ts.scale = self.scale
        # Disable the default tip names config
        ts.show_leaf_name = False
        ts.show_branch_support = show_branch_support

        # ts.branch_vertical_margin = 20
        ts.show_scale = show_scale
        cur_max_color = max(v.color for v in self.tree.traverse())
        current_colors = colors[0:cur_max_color + 1]

        for i, (label, color_) in enumerate(zip(legend_labels, current_colors)):
            if color_ not in used_colors: continue
            rf = RectFace(20 * legend_scale, 16 * legend_scale, color_, color_)
            rf.inner_border.width = 1
            rf.margin_right = 14
            rf.margin_left = 14

            tf = TextFace(label, fsize=26 * legend_scale)
            tf.margin_right = 14

            ts.legend.add_face(rf, column=0)
            ts.legend.add_face(tf, column=1)

        if neighbours:
            old_tree = self.tree.copy()
            self.draw_neighbours(neighbours, neighbours_block)

        self.tree.render(file, w=1000, tree_style=ts)

        if neighbours:
            self.tree = old_tree

    def get_all_leafs(self):
        return {node.name for node in self.tree.get_leaves()}

    def count_innovations_fitch(self, leaf_colors, count_second_color=True):
        def assign_colorset_feature(v):
            if v.is_leaf():
                v.add_features(colorset={leaf_colors[v.name]}, color=leaf_colors[v.name])
            else:
                try:
                    child1, child2 = v.children
                except ValueError:
                    print(v.children)
                    raise ValueError('Tree must me binary')
                cs1 = assign_colorset_feature(child1)
                cs2 = assign_colorset_feature(child2)
                v.add_features(colorset=(cs1 & cs2) if len(cs1 & cs2) > 0 else cs1 | cs2)

            return v.colorset

        def chose_color(colorset):
            return sorted(colorset, key=lambda c: color_counter[c], reverse=True)[0]

        def down_to_leaves(v, color):
            if v.is_leaf(): return
            v.add_features(color=color if color in v.colorset else chose_color(v.colorset))
            for child in v.children:
                down_to_leaves(child, v.color)

        def count_innovations(v, innovations, insertions_deltitions):
            for child in v.children:
                if v.color != child.color and not (not count_second_color and (v.color == 2) or (child.color == 2)):
                    innovations[child.color].append(child)
                    insertions_deltitions[child.color - v.color].append(child)
                count_innovations(child, innovations, self.insertions_deltitions)

        color_counter = Counter(leaf_colors.values())
        # print(color_counter)

        # get colorsets for internal nodes
        root = self.tree.get_tree_root()
        assign_colorset_feature(root)

        # get color for internal nodes
        root_color = chose_color(root.colorset)
        down_to_leaves(root, root_color)

        # get inconsistent colors
        self.innovations = defaultdict(list)
        self.insertions_deltitions = defaultdict(list)
        count_innovations(root, self.innovations, self.insertions_deltitions)

    def count_parallel_rearrangements(self, skip_grey):
        score, count, count_all = 0, 0, 0
        for color, nodes in self.innovations.items():
            if len(nodes) <= 1 or (skip_grey and color == 1): continue
            count += 1
            count_all += len(nodes)
            for n1, n2 in combinations(nodes, 2):
                score += n1.get_distance(n2)
        return score, count, count_all

    def count_parallel_breakpoints(self):
        count = sum(map(len, self.innovations.values()))
        score = sum(
            n1.get_distance(n2) for n1, n2 in combinations((n for ns in self.innovations.values() for n in ns), 2))
        return score, count

    def draw_coloring(self, file):
        for node in self.tree.traverse():
            node.img_style['bgcolor'] = self.colors[node.color]
        ts = TreeStyle()
        ts.show_leaf_name = False
        self.tree.render(file, w=1000, tree_style=ts)

    def prune(self, ls):
        self.tree.prune(list(ls))



import csv
import os

from statistics import mean

from collections import defaultdict
from textwrap import wrap


def get_characters_stats_unbalanced(blocks, characters, tree_holder):
    ans = []
    for block, genome_colors in zip(blocks, characters):
        tree_holder.count_innovations_fitch(genome_colors)

        ins_del_dict = tree_holder.insertions_deltitions
        insetions = sum(len(nodes) for change, nodes in ins_del_dict.items() if change > 0)
        delitions = sum(len(nodes) for change, nodes in ins_del_dict.items() if change < 0)

        score_rear, count_rear, count_all_rear = tree_holder.count_parallel_rearrangements(skip_grey=False)
        mean_copies = mean(genome_colors.values())
        ans.append([block, score_rear, count_rear, count_all_rear, mean_copies, count_all_rear <= 1, insetions, delitions])

    return ans


def write_stats_csv_unbalanced(stats, cls, stats_file):
    rows = [['block', 'cluster', 'parallel_rear_score', 'number_of_inconsistent_colors', 'number_of_parallel_events',
             'mean_copies', 'tree_consistent', 'insertions_count', 'deletions_count']] + \
           [stat[0:1] + [cl] + stat[1:] for stat, cl in zip(stats, cls)]
    with open(stats_file, 'w') as f:
        wtr = csv.writer(f)
        wtr.writerows(rows)


def call_unique_characters_one_cluster(cur_characters, cur_stats):
    count_blocks = defaultdict(list)
    for char, stats in zip(cur_characters, cur_stats):
        block = stats[0]
        count_blocks[frozenset(char.items())].append(block)
    return count_blocks


def call_unique_characters(cls_chars, cls_stats):
    return [call_unique_characters_one_cluster(c, s) for c, s in zip(cls_chars, cls_stats)]


def split_filenames(blocks, width=240):
    files = wrap(' '.join(map(str, blocks)), width, break_long_words=False)
    return [file.replace(' ', ',') for file in files]

def full_filenames(files, suffix):
    return [f'block{"s" if file.find(",") != -1 else ""}_' + file + suffix for file in files]


def write_characters_csv_unbalanced(unique_chars_list, folder):
    fill_length = len(str(len(unique_chars_list)))
    for cl, unique_chars in enumerate(unique_chars_list):
        cl_folder = folder + f'cluster_{str(cl).zfill(fill_length)}/'
        os.makedirs(cl_folder, exist_ok=True)
        for unique_char, unique_count_blocks in unique_chars.items():
            # unfrozen = dict(unique_char)
            rows = [['strain', 'character_state', 'character_state_annotation']] + \
                   [[strain, color, f'{color} copies'] for strain, color in unique_char]

            files = split_filenames(unique_count_blocks)
            for file in full_filenames(files, '.csv'):
                with open(cl_folder + file, 'w') as f:
                    wtr = csv.writer(f)
                    wtr.writerows(rows)

def write_trees_unbalanced(unique_chars_list, folder, show_branch_support, tree_holder, colors):
    fill_length = len(str(len(unique_chars_list)))
    for cl, unique_chars in enumerate(unique_chars_list):
        cl_folder = folder + f'cluster_{str(cl).zfill(fill_length)}/'
        os.makedirs(cl_folder, exist_ok=True)

        for unique_char, unique_count_blocks in unique_chars.items():
            unfrozen = dict(unique_char)
            tree_holder.count_innovations_fitch(unfrozen)

            labels = [f'{i}{"+" if i == len(colors) - 1 else ""} copies'
                      for i in range(max(unfrozen.values()) + 1)]
            files = split_filenames(unique_count_blocks)
            for file in full_filenames(files, '.pdf'):
                tree_holder.draw(cl_folder + file, legend_labels=labels, show_branch_support=show_branch_support, colors=colors)


import os

def write_trees_neightbours(blocks, ub_characters, neighbours, folder, show_branch_support, tree_holder, colors):
    for block, char in zip(blocks, ub_characters):
        labels = [f'{i}{"+" if i == len(colors) - 1 else ""} copies'
                  for i in range(max(char.values()) + 1)]

        file = folder + f'block_{block}.pdf'
        tree_holder.count_innovations_fitch(char)

        ns_block = neighbours[block]
        if len(set([frozenset(ns) for ns in ns_block.values()])) == 1: continue

        tree_holder.draw(file, legend_labels=labels, show_branch_support=show_branch_support, colors=colors,
                         neighbours=ns_block, neighbours_block=block, mode='r')


import numpy as np
import pandas as pd

from io import StringIO

COLUMNS = ["block", "species", "chr", "chr_beg", "chr_end", "orientation"]
BLOCKS_SEPARATOR = '-' * 80


def block_coords_to_infercars(in_file, out_file):
    lines = open(in_file).read().split('\n')[:-1]
    ls = np.array(lines)
    bs = np.split(ls, np.where(ls == BLOCKS_SEPARATOR)[0])

    # names of chromosomes
    df_names = pd.read_csv(StringIO('\n'.join(bs[0])), sep='\t')
    chr_names = {}
    for index, row in df_names.iterrows():
        chr_names[row['Seq_id']] = row['Description']

    # blocks data
    with open(out_file, 'w') as f:
        for b in bs[1:-1]:
            block = b[1].split('#')[1]
            df_block = pd.read_csv(StringIO('\n'.join(b[2:])), sep='\t')

            print(f'>{block}', file=f)
            for index, row in df_block.iterrows():
                chr_name, start, end, strand = chr_names[row['Seq_id']], row['Start'], row['End'], row['Strand']
                print(chr_name + '.1:' + (f'{start}-{end}' if strand == '+' else f'{end}-{start}') + ' ' + strand, file=f)
            print(file=f)


from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / 'README.md').read_text(encoding='utf-8')

setup(
    name='PaReBrick',
    version='0.5.7',
    description='A bioinf tool for finding genome rearrangements in bacterial genomes',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/ctlab/parallel-rearrangements',
    author='Alexey Zabelkin',
    author_email='a.zabelkin@itmo.ru',
    classifiers=[
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3 :: Only',
    ],
    keywords='genome rearrangements, phylogenetic trees, non-convex characters, synteny blocks, phylogenetic tree, '
             'pattern consistency',
    package_dir={'parebrick': 'parebrick'},
    packages=['parebrick', 'parebrick.characters', 'parebrick.clustering', 'parebrick.tree', 'parebrick.utils',
              'parebrick.utils.data'],
    python_requires='>=3.6, <3.9',
    install_requires=
        ['PyQt5', # for ete3 working properly
         'ete3', # phylogenetic trees
         'scikit-learn', # clustering characters
         'seaborn>=0.11.0', # for drawer module
         'bg'], # breakpoint graphs
    entry_points={
        'console_scripts': [
            'PaReBrick=parebrick.main:main',
            'parebrick=parebrick.main:main',
            'PaReBrick-Charts=parebrick.drawer:main',
            'parebrick-charts=parebrick.drawer:main',
        ],
    },
)


from collections import defaultdict

import numpy as np

import csv
import os


def get_characters_which_chr(blocks_df):
    def make_label(chrs):
        return 'chromosomes: ' + ', '.join(map(str, sorted(chrs)))

    characters = []

    which_chrs = defaultdict(lambda: defaultdict(list))
    block_lens = defaultdict(list)

    for (block, genome), df_block_genome in blocks_df.groupby(['block', 'species']):
        which_chrs[block][genome] = df_block_genome.chr.tolist()
        block_lens[block].extend(df_block_genome.chr_end - df_block_genome.chr_beg)

    for block in sorted(blocks_df.block.unique()):
        used_chroms = set()
        genome_colors = defaultdict(int)

        labels = ['no copies'] \
                 + sorted(list(set(make_label(chrs) for chrs in which_chrs[block].values())))

        for genome, chromosomes in which_chrs[block].items():
            label = make_label(chromosomes)
            used_chroms.update(chromosomes)

            color = labels.index(label)
            genome_colors[genome] = color

        multichromo = ';'.join(map(str, sorted(list(used_chroms))))
        mean_length = np.mean(block_lens[block])
        std_length = np.std(block_lens[block])

        characters.append((block, multichromo, mean_length, std_length, genome_colors, labels))

    return characters


def get_characters_stats_which_chr(characters, tree_holder):
    ans = []
    for block, multichromo, mean_length, std_length, genome_colors, labels in characters:
        tree_holder.count_innovations_fitch(genome_colors)

        score_rear, count_rear, count_all_rear = tree_holder.count_parallel_rearrangements(skip_grey=False)
        ans.append([block, multichromo, mean_length, std_length, score_rear, count_rear, count_all_rear, count_all_rear <= 1])

    return ans


def write_stats_csv_which_chr(stats, stats_file):
    rows = [['block', 'chroms', 'mean_length', 'std_length', 'parallel_rear_score', 'number_of_inconsistent_colors',
             'number_of_parallel_events', 'tree_consistent']] + stats
    with open(stats_file, 'w') as f:
        wtr = csv.writer(f)
        wtr.writerows(rows)


def write_characters_csv_which_chr(characters, folder):
    os.makedirs(folder, exist_ok=True)

    for i, (block, _1, _2, _3, genome_colors, labels) in enumerate(characters):
        rows = [['strain', 'character_state', 'character_state_annotation']] + \
               [[strain, color, labels[color]] for strain, color in genome_colors.items()]

        with open(folder + f'block_{block}.csv', 'w') as f:
            wtr = csv.writer(f)
            wtr.writerows(rows)


def write_trees_which_chr(characters, folder, show_branch_support, tree_holder, colors):
    os.makedirs(folder, exist_ok=True)

    for i, (block, _1, _2, _3, genome_colors, labels) in enumerate(characters):
        tree_holder.count_innovations_fitch(genome_colors)
        tree_holder.draw(folder + f'block_{block}.pdf', legend_labels=labels,
                         show_branch_support=show_branch_support, colors=colors)


import numpy as np

from collections import defaultdict


def distance_between_blocks_distribution(df_blocks):
    ds = []
    for sp, df_sp in df_blocks.groupby('species'):
        df_sp = df_sp.sort_values(by=['chr_beg'])
        ds += (start_ - end_ for start_, end_ in zip(df_sp['chr_beg'][1:], df_sp['chr_end']))
    return ds


def distance_between_blocks_dict(df_blocks, genome_length, allowed_blocks=None):
    distances = defaultdict(lambda: defaultdict(lambda: 1e12))

    if allowed_blocks is None:
        df = df_blocks.copy()
    else:
        df = df_blocks.loc[df_blocks['block'].isin(allowed_blocks)].copy()

    df['real_beg'] = df.apply(lambda row: row.chr_beg if row.orientation == '+' else row.chr_end, axis=1)
    df['real_end'] = df.apply(lambda row: row.chr_beg if row.orientation == '-' else row.chr_end, axis=1)

    for strain, df_strain in df.groupby('species'):
        for chr, df_strain_chr in df_strain.groupby('chr'):
            l = genome_length[strain + '.' + chr]

            xs = df_strain_chr[['block', 'real_beg', 'real_end']].to_numpy()

            m, n = xs.shape

            combs = np.zeros((m, m, 2 * n), dtype=int)
            combs[:, :, :n] = xs[:, None, :]
            combs[:, :, n:] = xs
            combs.shape = (m * m, -1)

            combs = combs[np.where((combs[:, 3] - combs[:, 0]) > 0), :][0]

            for b1, st1, end1, b2, st2, end2 in combs:
                if b1 >= b2: continue
                distances[(b1, b2)][strain] = min(distances[(b1, b2)][strain],
                                                  abs(st1 - st2),
                                                  l - abs(st1 - st2),
                                                  abs(st1 - end2),
                                                  l - abs(st1 - end2),
                                                  abs(end1 - st2),
                                                  l - abs(end1 - st2),
                                                  abs(end1 - end2),
                                                  l - abs(end1 - end2),
                                                  )

    return distances


def check_stats_stains(tree, block_genomes, logger):
    tree_genomes = tree.get_all_leafs()

    logger.info(f'Strains in blocks file count: {len(block_genomes)}')
    logger.info(f'Strains in tree leafs count: {len(tree_genomes)}')
    logger.info(f'Intersected strains count: {len(block_genomes & tree_genomes)}')

    if not len(block_genomes & tree_genomes) == len(tree_genomes) == len(block_genomes):
        if len(block_genomes & tree_genomes) == 0:
            logger.error(f'Tree genomes: {tree_genomes}')
            logger.error(f'Blocks genomes: {block_genomes}')
            logger.error('Seems like strains in block files and in tree leafs have different ids, intersection is empty.')
            raise ValueError(
                'Seems like strains in block files and in tree leafs have different ids, intersection is empty.')
        left_tree = tree_genomes - (block_genomes & tree_genomes)
        if len(left_tree) > 0:
            logger.warning('PRUNING TREE')
            logger.warning(f'Those strains are thrown away from tree because there were not found in blocks data: {", ".join(left_tree)}')
            tree.prune(block_genomes & tree_genomes)
        left_blocks = block_genomes - (block_genomes & tree_genomes)
        if len(left_blocks) > 0:
            logger.warning('FILTERING GENOMES IN BLOCKS')
            logger.warning(f'Those strains are thrown away from blocks because there were not found in tree leafs: {", ".join(left_blocks)}')

    return block_genomes & tree_genomes


def get_coverages(df, genome_lengths):
    return [sum(df_strain['chr_end'] - df_strain['chr_beg']) / genome_lengths[strain]
            for strain, df_strain in df.groupby('species')]


def get_mean_coverage(df, genome_lengths):
    return np.mean(get_coverages(df, genome_lengths))



from bg.grimm import GRIMMReader
from bg.tree import BGTree
from bg.genome import BGGenome
from bg.vertices import BGVertex


import numpy as np

import os
import csv

from collections import Counter

get_colors_by_edge = lambda e: e.multicolor.multicolors

def white_proportion(colors):
    return np.mean(list(map(lambda c: c == 0, colors)))

def get_character_by_edge(bg, edge, genomes, neighbour_index):
    def get_neighbour_with_genome(v, genome):
        return neighbour_index[(v, genome)]

    def get_genome_character_state_by_edge(genome):
        if cnt[genome] == 1:
            return 0
        else:
            v1, v2 = edge.vertex1.name, edge.vertex2.name
            if v1 > v2: v1, v2 = v2, v1

            try:
                v1_neighbour = get_neighbour_with_genome(v1, genome)
                v2_neighbour = get_neighbour_with_genome(v2, genome)
            except KeyError:
                return 2

            neighbour_edge = bg.get_edge_by_two_vertices(v1_neighbour, v2_neighbour)
            if neighbour_edge:
                neighbour_edge_genomes = get_colors_by_edge(neighbour_edge)
                neighbour_edge_genomes_count = sum(neighbour_edge_genomes.values())

                neighbour_edge_genomes_names = set(g for g in genomes if neighbour_edge_genomes[BGGenome(g)])
                white_genomes_at_neighbour_edge = len(neighbour_edge_genomes_names & white_genomes)

                if white_genomes_at_neighbour_edge < 1:
                # if neighbour_edge_genomes_count <= len(genomes) // 2:
                    return 1

                pair = (v1_neighbour, v2_neighbour)
                if pair not in possible_edges:
                    possible_edges.append(pair)
                return 3 + possible_edges.index(pair)
            else:
                return 1

    cnt = Counter([genome.name
                   for genome, count in get_colors_by_edge(edge).items()
                   for _ in range(count)])
    possible_edges = []
    white_genomes = set(g for g in genomes if BGGenome(g))
    return {genome: get_genome_character_state_by_edge(genome) for genome in genomes}, possible_edges

def construct_vertex_genome_index(bg):
    neighbour_index = {}
    for v in bg.bg:
        for edge in bg.get_edges_by_vertex(v):
            colors = get_colors_by_edge(edge)
            for color in colors:
                neighbour_index[(str(v), color.name)] = edge.vertex2

    return neighbour_index

def get_characters_balanced(grimm_file, genomes, logger):
    bg = GRIMMReader.get_breakpoint_graph(open(grimm_file))
    logger.info('Breakpoint graph parsed')

    logger.info(f'Edges in breakpoint graph: {len(list(bg.edges()))}')

    characters = []

    # consistency_checker = TreeConsistencyChecker(tree_file)
    for i, component_bg in enumerate(bg.connected_components_subgraphs()):
        nodes_len = len(list(component_bg.nodes()))
        if nodes_len == 2: continue

        logger.info(f'Getting characters from breakpoint graph component, size={len(component_bg.bg)}')

        neighbour_index = construct_vertex_genome_index(component_bg)

        for edge in component_bg.edges():
            v1, v2 = edge.vertex1.name, edge.vertex2.name
            if v1 > v2: v1, v2 = v2, v1

            genome_colors, neighbour_edges = get_character_by_edge(component_bg, edge, genomes, neighbour_index)
            if white_proportion(genome_colors.values()) < 0.33: continue

            labels = ['adjacency exists', 'complex break of adjacency', 'some block is not presented'] + \
                     [f'inversion with {v1n}-{v2n}' for (v1n, v2n) in neighbour_edges]

            characters.append((v1, v2, genome_colors, labels))

    return characters

def get_characters_stats_balanced(characters, tree_holder, distance_between_blocks):
    ans = []
    for v1, v2, genome_colors, labels in characters:
        tree_holder.count_innovations_fitch(genome_colors, count_second_color=False)

        b1, b2 = int(v1[:-1]), int(v2[:-1])
        if b1 > b2: b1, b2 = b2, b1

        score_rear, count_rear, count_all_rear = tree_holder.count_parallel_rearrangements(skip_grey=True)
        score_break, count_break = tree_holder.count_parallel_breakpoints()

        white_strains = [strain for strain, color in genome_colors.items() if color == 0]
        mean_break_length = np.mean([distance_between_blocks[(b1, b2)][strain] for strain in white_strains])

        ans.append([f'{v1}-{v2}', int(mean_break_length), score_rear, count_rear, count_all_rear, score_break, count_break,
                    count_break <= 1])

    return ans

def write_stats_csv_balanced(stats, stats_file):
    rows = [['id', 'adjacency', 'mean_break_length_nucleotide', 'parallel_rear_score', 'number_of_inconsistent_colors',
            'number_of_parallel_events', 'parallel_break_score', 'number_of_parallel_breaks', 'tree_consistent']] + \
          [[i+1] + stat for i, stat in enumerate(stats)]
    with open(stats_file, 'w') as f:
        wtr = csv.writer(f)
        wtr.writerows(rows)

def write_characters_csv_balanced(characters, folder):
    os.makedirs(folder, exist_ok=True)
    fill_length = len(str(len(characters)))

    for i, (v1, v2, genome_colors, labels) in enumerate(characters):
        rows = [['strain', 'character_state', 'character_state_annotation']] + \
               [[strain, color, labels[color]] for strain, color in genome_colors.items()]

        with open(folder + f'id_{str(i + 1).zfill(fill_length)}_edge_{v1}-{v2}.csv', 'w') as f:
            wtr = csv.writer(f)
            wtr.writerows(rows)

def write_trees_balanced(characters, folder, show_branch_support, tree_holder, colors):
    os.makedirs(folder, exist_ok=True)
    fill_length = len(str(len(characters)))

    for i, (v1, v2, genome_colors, labels) in enumerate(characters):
        tree_holder.count_innovations_fitch(genome_colors)
        tree_holder.draw(folder + f'id_{str(i + 1).zfill(fill_length)}_edge_{v1}-{v2}.pdf', legend_labels=labels,
                         show_branch_support=show_branch_support, colors=colors)


import numpy as np


def dist_matrix_similarity(characters):
    n = len(characters)
    arr = np.zeros((n, n))
    for i1, char1 in enumerate(characters):
        for i2, char2 in enumerate(characters):
            if i1 >= i2: continue
            arr[i1, i2] = arr[i2, i1] = \
                sum(abs(char1[genome] - char2[genome]) for genome in char1.keys() | char2.keys())

    return arr


def dist_matrix_proximity(stats, distance_between_blocks, max_length, percentile):
    n = len(stats)
    arr = np.zeros((n, n))
    for i1, stats1 in enumerate(stats):
        for i2, stats2 in enumerate(stats):
            block1, block2 = stats1[0], stats2[0]
            if i1 >= i2: continue
            b1, b2 = block1, block2
            if b1 > b2: b1, b2 = b2, b1
            ds = distance_between_blocks[(b1, b2)].values()
            arr[i1, i2] = arr[i2, i1] = np.percentile(list(ds), percentile) if len(ds) > 0 else max_length

    return arr
